/* ReSCAL - genesis entry
 *
 * Copyright (C) 2011-2012
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define _MAIN_

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "defs.h"
#include "macros.h"
#include "cells.h"
#include "genesis.h"
#include "param.h"
#include "format.h"
#include "templates.h"


int prog=PROG_TOOL;
char* comp_date=__DATE__;
int  H, L, D, HL, HLD;        // size of cellular space
Cell *TE=NULL;          // cellular space
double csp_time=0.0;
//int  node = -1;            // node number (with PARALLEL option)
char *model_name = NULL; //name of model
//char *bin_filename = NULL; //name of BIN file
//char *csp_filename = NULL; //name of CSP file
char opt_bin = 0;   //raw binary format
int graine = 68374; //random seed
int use_genesis = 1;

extern CSP_Template t_templates[]; //template types
extern int nb_template_types; ////number of template types
extern CSP_Template csp_templates[]; //array of CSP templates from parameter file
extern int nb_templates; //number of templates in the array from parameter file

char *boundary_str = NULL;
int boundary = BC_PERIODIC;  //boundary conditions

char *rot_map = NULL;
int rot_mode = 0;
int L_border = 0;
int H_border = 0;
int D_border = 0;

double drand48();
void  srand48();


void genesis_parse()
{
  if (model_name && strcmp(model_name, MOD_NAME)){
    WarnPrintf("ERROR: Wrong model name %s in parameter file. Check model in defs.h and rebuild all.\n", model_name);
    exit(-2);
  }

  //if (csp_template_str){
  //parse_template_array();
  //}

  if (boundary_str){
    if (!strcmp(boundary_str, "PERIODIC")){
#ifdef CYCLAGE_HOR
      boundary = BC_PERIODIC;
#else
      ErrPrintf("Incorrect value for Boundary : %s, please define CYCLAGE_HOR in defs.h and recompile\n", boundary_str);
      exit(-2);
#endif
    }
    else if (!strcmp(boundary_str, "OPEN")){
      boundary = BC_OPEN;
    }
    else if (!strcmp(boundary_str, "OUT")){
      boundary = BC_OUT;
    }
    else if (!strcmp(boundary_str, "CLOSED") || !strcmp(boundary_str, "CLOSE")){
      boundary = BC_CLOSED;
    }
    else if (!strcmp(boundary_str, "REINJECTION")){
      boundary = BC_REINJECTION;
    }
    else{
      ErrPrintf("Incorrect value for Boundary : %s\n", boundary_str);
      exit(-2);
    }
    //pbc_mode = (boundary == BC_PERIODIC);
  }
}


///////////////////////////////////////////////////////////////////////////////
/// Custom setup of a single cell
///////////////////////////////////////////////////////////////////////////////
void genesis_cell_custom(Cell *aux, int i, int j, int k)
{
  /// CSP_CUSTOM: this part can be freely customized to set the initial cellular space

  int H0 = H-2;
  int Ld2 = L/2;
  int Dd2 = D/2;
  int Hd2 = H/2;
  int Ld3 = (int) L/3;
  int Ld4 = (int) L/4;
  int Ld8 = (int) L/8;
  int Dd3 = (int) D/3;
  int Dd4 = (int) D/4;
  int Hd3 = (int) H/3;
  int Hd4 = (int) H/4;
  float Ldx= Ld2-0.5;
  float Ldy= Dd2-0.5;

/*****************************************************************************/
/********************************* DUN model *********************************/
/*****************************************************************************/
#ifdef MODEL_DUN

  //int ll1 = (int) L*0.2; //L*0.15; //L*0.5;//L*0.1; //10; //0.1 //0.22; //Ld3
  //int ll2 = (int) L*0.2; //0.22; //Ld3
  //int l0 = (int) L/8; //L*0.1; //(L-ll1-ll2)*0.4;//25; //0.45
  //int l1 = (int) l0+ll1*1.3; //pour 2 triangles
  //int l2 = (int)(L-ll1-ll2)*0.25; //0.45
  //int hh = (int) (H/6);//(100/3); //8+(100/3); //(H/3);//(60/2); //H/6; //H/2;
  //int hh1 = hh/1.5;
  //int hh2 = hh;
  //int lc = D; //1; //400; //Ld4; //Ld2; //L/10; //Ld4; //largeur du couloir
  //int lcone = lc*0.6; //lc*0.6; //lc/2; //lc*0.6; //lc*2/3;  //Ld2*0.8; //largeur du cone
  //int jcone = H0-lcone/2; //sommet du cone
  //int hcone = Hd3; //Hd2;//2*Hd3; //H-10; //hauteur du cone
  //int h0 = (int) (H-hh)/2;
  //int h2 = H-h0;
  //int rcyl = Ld2-10; //rayon cylindre
  //float periode = 10.0; //periode des ondulations
  //float pente = 2.0;
  //float Ldx = Ld2-0.5; //Ld2-0.5; //Ld3;//Ld4;
  //float Ldy = /*0.66**/Dd2-0.5; //Ld4; //(int) L/2;
  //float Ldz = Hd2;
  //di = i - Ldx;
  //dj = j - Ldz;
  //dk = k - Ldy; // + Ld4;
  //periode = 10.0+(k-(Dd2-lc/2))/2.0/2.0;
  //periode += 0.5*drand48();

  /*** GR cells ***/
  //rc = 0.5*lcone*(j-(H-hcone))/(hcone-1);
  if (0) // no sand
  //if ((di*di + dk*dk < (j-(H-hcone))*(j-(H-hcone))) && (j > (H-hcone))) //cone de hauteur hcone
  //if ((di*di + dk*dk < 4*(j-2*Hd3)*(j-2*Hd3)) && (j > 2*Hd3)) //cone de hauteur Hd3 et de rayon 2*Hd3
  //if ((di*di + dk*dk < 2.5*(j-Hd3)*(j-Hd3)) && (j > Hd3)) //cone de hauteur 2*Hd3 pour les dunes etoiles
  //if ((di*di + dk*dk < (j-jcone)*(j-jcone)) && (j > Hd2)) //cone tronque
  //if (((di*di + dk*dk < (j-jcone)*(j-jcone)) && (j > jcone)) || (j > H-10)) //cone + couche uniforme
  //if ((j > 12) && (di*di + dk*dk < 400)) // disque
  //if ((j > H*0.65) && (fabs(di) < Ld8) && (fabs(dk) < Ld8)) // parallelepipede rectangle
  //if ((j > Hd2) && ((di+dk)*(di+dk) < 400) && (dk*dk < 400)) // parallelepipede 'tordu'
  //if (j + (i/20)> H-10) //pente EO
  //if (j + k*pente > H-20) //pente NS
  //if (j + i > H) //pente EO
  //if (j > H-10) //couche uniforme
  //if (j>H-8-(drand48()-0.90)) //couche aleaoire
  //if ((j > H-4) && (k>=Dd4) && (k<D-Dd4)) //couche uniforme est-ouest
  //if ((j > H-5) && (di*di + dk*dk < rcyl*rcyl)) //couche uniforme circulaire
  //if (/*(i>1) &&*/ (j > H-Hd2*drand48())) //couche non-uniforme
  //if ((j > H-2) || ((j == H-2) && i < L*0.5 && (drand48()>0.75/*0.5*/))) //couche non-uniforme
  //if (j > Hd2-5*cos(10*i*PI/L)) //ondulations
  //if (j > Hd2-2*sin(i*2*PI/periode)) //ondulations variables
  //if (((i>=l0) && (i<l0+1/*l0*/) && (j>H-hh)) || (j>=H-1)) //rectangle
  //if (((i>=l0) && (i<l0+1) && (j>H-hh)) || (j< (H-hh-10)*(Ld2-k)/lc) || (j>=H-1)) //rectangle avec toit incline
  //if (((i>=l0) && (i<l0+ll1) && ((i-l0)-((H0-j/*+drand48()*5*/)*ll1/hh) > 0))
  //      || (j>=H-1)) //triangle
  //if (((i>=l0) && (i<l0+ll1) && ((i-l0)-((H0-j/*+drand48()*5*/)*ll1/hh) > 0))
  //           || ((i>=l0+ll1) && ((i-l0-ll1)-(hh-(H0-j))/5 < 0))
  //           || (j>=H0)) //triangle + triangle inverse
  //if (((i>=l0) && (i<l0+ll1) && ((i-l0)-((H-j)*ll1/hh1) > 0))
  //      || ((i>=l1) && (i<l1+ll2) && ((i-l1)-((H-j)*ll2/hh2) > 0))
  //      || (j>=H-1)) //2 triangles
  /*if (((i>=l0) && (i<=l0+ll1) && ((i-l0)-((H-j)*ll1/hh) > 0)) || (j< H*((k<Ld2) ? Ld2-k : k-Ld2)/Ld2)
    || (j>=H-1)) //triangle avec toit incline*/
                  /*else if (((i>=l0) && (i<=l0+ll1) && ((i-l0)-((H-j)*ll1/hh) > 0)) || (j< Hd2*((k<Ld2) ? Ld2-k : k-Ld2)/ll)
    || (j>=H-1)) //triangle avec toit en biseau*/
  //if (((i>=l0) && (i<=l0+ll1) && ((i-l0)-((H-j)*ll1/hh) > 0)) || (j< (H-hh-10)*(Ld2-k)/lc) || (j>=H-1)) //triangle avec toit incline
  /*if (((i>=l0) && (i<l0+ll1) && ((i-l0)-((H-j)*ll1/hh) > 0))
    || ((i>=l0+ll1) && (i<L-ll2-l2) && (j>=hh))
    || ((i>=L-ll2-l2) && (i<L-l2) && ((L-l2-i)-((H-j)*ll2/hh) > 0))
    || (j>=H-1)) //trapeze*/
  //if (di*di + dj*dj + dk*dk < Hd4*Hd4) // sphere
  //if ((di*di + dj*dj + dk*dk < Hd4*Hd4) || ((di+50)*(di+50) + (dj+30)*(dj+30) + dk*dk < Hd4*Hd4/4)) // 2 spheres
  //if (j >H0-3*(1+cos(10*i*k*PI/L))) //test mh
  //if (j >H0-3*(1+cos(10*i*k*PI/L/D))) //test mh
  //if (j >H0-3*(0+cos(10*i*PI/L)-cos(10*k*PI/D))) // champs de bosses
  //if ((((l1=i%(l0*2))<l0) && (j>H0-Hd3+hh/2-(hh*l1/l0))) || ((l1>=l0) && (j>H0-Hd3+hh/2-(hh*(2*l0-l1)/l0)))) // barres triangulaires periodiques
  {
    aux->celltype = GR;
  }

  /*** DUM cells ***/
  if (0)  // no DUM cells added
  //if (j > H-2) //plateau
  //if ((j > H0) || (j==0)) //plateau + plafond
  //if ((j > H0-round(drand48())) || (j==0)) //plateau rugueux + plafond
  //if ((j > H0) || (j==0) || ((i==Ld3) && (j>=Hd2))) //plateau + plafond + mur
  //if ((j > H0) || (j==0) || (k>=D-5) || (k<5)) //plateau + plafond + couloir
  //if ((j > H0) || (j<4)) //plateau + plafond epais (4)
  //if ((k < Dd2-lc/2) || (k > Dd2+lc/2) || (k & 1) || (j > H0) || (j==0)) //plateau + plafond + couloir + peigne
  //if ((j > H0-5*(1+cos(10*i*PI/L))) || (j==0)) //plateau ondule + plafond
  //if ((j > H0-5*(1+cos(10*i*PI/L))) || (j < 4)) //plateau ondule + plafond epais (4)
  //if (((i == L-1) && (j>H-L))) //couloir + mur
  //if ((j > H-2) && (di*di + dk*dk < Ld2*Ld2)) // disque + plateau + plafond
  //if (((j > hgrain) && (di*di + dk*dk >= rcyl*rcyl)) || (j > H0) || (j==0)) // plateau + trou circulaire + plafond
  //if (j + (i/20)> H-2) //pente EO
  //if ((j + (i/20)> H-2) || (j==0)) //pente EO + plafond
  //if (j + k*pente > H+20) //pente NS
    if (aux->celltype != IN){
      aux->celltype   = DUM;
#ifdef CELL_COLOR
      aux->color = 0;
#endif
    }

  /*** IN cells ***/
  if (0/*boundary == BC_REINJECTION)*/){
    //if ((k == Ld2) && (i == Ld2) && (j == 0)) //source de grains en haut au centre
    if (/*(node == 0) &&*/ (i == 5) && (j == 0)) //source de grains en haut a gauche
    //if ((j == 0) && (di*di + dk*dk < 5*5)) ////source de grains en disque
    //if (((i == 40) || (i == 41)) && (j > H0)) //source de grains nord-sud au sol
    /*float Ldx1 = 100-0.5;//-Ld3*cos(PI/4);
    float Ldy1 = Dd2-0.5;//-Ld3*sin(PI/4);
    if ((j > H0) && ((i-Ldx1)*(i-Ldx1) + (k-Ldy1)*(k-Ldy1) <= 20*20)) ////source de grains en disque au sol
      aux->celltype = IN;*/
    /*float Ldx1 = Ld2-0.5-Ld3*cos(PI/3)+(L/10)*sin(PI/3);
    float Ldy1 = Dd2-0.5-Ld3*sin(PI/3)-(L/10)*cos(PI/3);
    float Ldx2 = Ld2-0.5-Ld3*cos(PI/3)-(L/10)*sin(PI/3);
    float Ldy2 = Dd2-0.5-Ld3*sin(PI/3)+(L/10)*cos(PI/3);
    if ((j > H0) && (((i-Ldx1)*(i-Ldx1) + (k-Ldy1)*(k-Ldy1) <= 25)
                 || ((i-Ldx2)*(i-Ldx2) + (k-Ldy2)*(k-Ldy2) <= 25))) ////2 sources de grains en disque au sol*/
    //if ((j > H0) && ((di*di + dk*dk >= Ld3*Ld3) && ((di+cos(PI/3))*(di+cos(PI/3)) + (dk+sin(PI/3))*(dk+sin(PI/3)) < Ld3*Ld3))) ////source de grains en demi-cercle au sol, sur la table tournante, avec un angle initial de Pi/3
    //if ((j == 0) && ((di*di + dk*dk >= rcyl*rcyl) && ((di+1)*(di+1) + dk*dk < rcyl*rcyl))) ////source de grains en demi-cercle
    //if ((j == 0) && (dk > -5) && (dk < 5) && ((di*di + dk*dk >= rcyl*rcyl) && ((di+1)*(di+1) + dk*dk < rcyl*rcyl))) ////source de grains en arc de cercle
    //if ((j == 0) && (((k >= i) && (i+k <= L) && (di*di + dk*dk >= rcyl*rcyl) && ((di+1)*(di+1) + dk*dk < rcyl*rcyl))
    //     || (((k <= i) || (i+k >= L)) && (i+1==round(Ld2-rcyl/sqrt(2)))))) //source de grains en quart de cercle + segments de droite
      aux->celltype = IN;
  }

  /*** OUT cells ***/
  //if ((j > H0) && (i>=L-5) && (i<L-1)) ////sortie de grains au sol pres du bord est
  //if ((j == H-1) && (di*di + dk*dk >= rcyl*rcyl)) ////sortie de grains en cercle
  //if ((j == H-1) && ((di>0) && ((di-1)*(di-1) + (dk-sgn(dk))*(dk-sgn(dk)) >= rcyl*rcyl))) ////sortie de grains en demi-cercle
    //aux->celltype = OUT;
  /*if ((j==H-1) && (k>=Dd3-2) && (k<Dd3+2)) ////sortie de grains entre les couloirs
    aux->celltype = OUT;
  if ((j==H-1) && (k>=2*Dd3-2) && (k<2*Dd3+2)) ////sortie de grains entre les couloirs
    aux->celltype = OUT;*/

#endif // MODEL_DUN

/*****************************************************************************/
/********************************* AVA model *********************************/
/*****************************************************************************/
#ifdef MODEL_AVA
  int hc = H*0.4;
  float di, dk;
  float tg = tan(PI/6);

  Ldx= Ld2;//5 + 0.5; //Ld4; //10+Hd2;
  Ldy= Dd2;//5 + 0.5; //Ld4; //(int) L/2;

  di = i - Ldx;
  dk = k - Ldy;
  if (di<0) di = -di;
  if (dk<0) dk = -dk;

  if ((di*di + dk*dk <= 25) && (j==0)) //source etendue de grains
  //else if ((((i-Ld3)*(i-Ld3) + dk*dk <= 25) || ((i-2*Ld3)*(i-2*Ld3) + dk*dk <= 25)) && (j==0)) //2 sources etendues de grains
  //else if ((j==0) && (di*di + dk*dk < 55*55) && ((di-1)*(di-1) + (dk-1)*(dk-1) >= 45*45)) //source circulaire de grains
    aux->celltype = IN;
  //else if ((di*di + dk*dk <= 25) && (j==H-2)) //sortie circulaire de grains
  //  aux->celltype = OUT;
  //else if ((j>H/4) && (di*di + dk*dk <= (Ld2-5)*(Ld2-5))) //cylindre
  //else if ((di*di + dk*dk < 2*(j-Hd3)*(j-Hd3)) && (j > Hd3)) //cone de hauteur 2*Hd3
  //else if ((di*di + dk*dk < (j-H+hc)*(j-H+hc)/(tg*tg)) && (j > H-hc)) //cone de hauteur hc et de pente tg
  //    aux->celltype   = DUM;
  //    aux->celltype   = GR;
#ifdef CELL_COLOR
    //aux->color = (drand48()<0.5) ? 1 : 0;
#endif
  //else if ((j == H-1) && (di*di + dk*dk >= (Ld2-5)*(Ld2-5))) ////sortie de grains en cercle
  //  aux->celltype = OUT;

#endif

/*****************************************************************************/
/********************************* CMB model *********************************/
/*****************************************************************************/
#ifdef MODEL_CMB
  /// two horizontal layers (solid on the top, magma at the bottom)
  if (j < Hd2) 
  //else if ((j < Hd2) || ((j < Hd2+20) && (i > Ld2-10) && (i < Ld2+10)))
    aux->celltype   = PLUS;
  else
    aux->celltype   = MOINS;
  //aux->celltype   = (int) floor(drand48()*3);
#endif // MODEL_CMB

/*****************************************************************************/
/********************************* ICB model *********************************/
/*****************************************************************************/
#ifdef MODEL_ICB
  /// two horizontal layers (mixed fluid on the top, solid core at the bottom)
  if (j < H*2/3)
    aux->celltype   = (drand48() < 0.5)?ZERO:MOINS;
  else
    aux->celltype   = PLUS;
  //aux->celltype   = (int) floor(drand48()*3);
#endif // MODEL_ICB

/*****************************************************************************/
/********************************* RIV model *********************************/
/*****************************************************************************/
#ifdef MODEL_RIV
  float niv = H*0.9;
  float hout = H-1; 
  int ll = Ld2;

  /*** TERRE cells ***/
  //if (j > niv)
  if (j > niv-((L-1-k)*0.9)) //pente de 0.9
  //if (j > niv-((L-1-k)/2)-((i<Ld2)?(Ld2-i):(i-Ld2))) //pente NS avec sillon
  //if (j > niv-((L-1-k)*0.8)+H*sin(5.0*i*6.28/L)/20) //pente de 0.8 + sinusoide
  //if ((j > niv-(k*0.9)) && (j > niv-((L-1-k)*0.9))) //triangle de pente 0.9
    aux->celltype = TERRE;
    //aux->celltype = (((int) floor(drand48()*100) == 0)) ? PIERRE : TERRE;

  /*** DUM cells ***/
  //if ((i==0) || (k==0) || (i==L-1) || (k==D-1)) aux->celltype = DUM;  //barriere infranchissable tout autour (modele ferme)
  if ((i==0) || (k==0) || (i==L-1) || ((k==D-1) && (j>=hout))) aux->celltype = DUM;  //barriere infranchissable tout autour et jusqu'a une certaine hauteur sur le bord sud
  //if ((i==0) || (i==L-1) || (((k==0) || (k==D-1)) && (j>=hout))) aux->celltype = DUM;  //barriere infranchissable tout autour et jusqu'a une certaine hauteur sur les bords sud et nord
  //if ((i==0) || (i==L-1) || (((k==0) || (k==D-1)) && ((j>=hout) || (i!=Ld2)))) aux->celltype = DUM;  //barriere infranchissable tout autour et jusqu'a une certaine hauteur au centre des bords sud et nord
  //if ((i<Ld2-(ll/2)) || (i>Ld2+(ll/2))) aux->celltype = DUM; //couloir

  /*** IN cells ***/
  if ((j==1) && (i==Ld2) && (k==1)) aux->celltype = IN; //cellule source au nord
#endif // MODEL_RIV

/*****************************************************************************/
/********************************* CRY model *********************************/
/*****************************************************************************/
#ifdef MODEL_CRY
  int H1 = H*0.4;
  int H2 = H*0.8;
  if ((j==H1) && (i<1) && (k<1))
    aux->celltype = IN;
  else
    aux->celltype = (j<H2) ? AIR : PLUS;
#endif // MODEL_CRY

/*****************************************************************************/
/********************************* DIF model *********************************/
/*****************************************************************************/
#ifdef MODEL_DIF
  /*** ONE cells ***/
  //if (i>=Ld2) aux->celltype = ONE; //vertical layer on the right half part
  
  /*** IN cells ***/
  if (i==0) aux->celltype = IN; //source on the west boundary
#endif // MODEL_DIF

/*****************************************************************************/
/********************************* D2G model *********************************/
/*****************************************************************************/
#ifdef MODEL_D2G
  if ((j==0) && (i==Ld2) && (k==Dd2))
    aux->celltype = IN;
  else if (j>=H-5)
    aux->celltype = TERRE;
  else
    aux->celltype = AIR;
#endif // MODEL_D2G

/*****************************************************************************/
/********************************* LIFE model ********************************/
/*****************************************************************************/
#ifdef MODEL_LIFE
  //aux->celltype = (drand48()<0.5) ? ALIVE : DEAD; //random
  //aux->celltype = (k & 1) ? ALIVE : DEAD; //horizontal lines
  //if ((i-Ld2)*(i-Ld2) + (k-Dd2)*(k-Dd2) < Ld3*Ld3) //disk
  //if (((i-Ld2)*(i-Ld2) + (k-Dd2)*(k-Dd2) < Ld4*Ld4) || (k & 1)) //disk + horizontal lines
  if ((i==Ld2) || (k & 1)) //1 vertical line + horizontal lines
  //if ((drand48()<0.01) || (k & 1)) //random cells + horizontal lines
    aux->celltype = ALIVE;
  else
    aux->celltype = DEAD;
#endif // MODEL_LIFE

}

///////////////////////////////////////////////////////////////////////////////
/// Final state of a cell (rules applied over all space, whatever the templates)
/// Genesis specific rules for bidisperse/colored granular cells, vegetation, etc.
///////////////////////////////////////////////////////////////////////////////
void genesis_cell_final(Cell *aux, int i, int j, int k)
{
#ifdef MODEL_DUN
  int H0 = H-2;

#ifdef CELL_COLOR
  float density = 0.5; //density of white grains

  /*** Color of cells ***/
  if (aux->celltype == GR){
    //aux->color = 1; //unformly colored
    aux->color = (drand48()<density) ? 1 : 0; //randomly colored
    //aux->color = (k > Dd2) ? 1 : 0; //colored area to the north
    //aux->color = (k == Dd2+1) && (j == H-20) && (i == 40) ? 1 : 0; //one colored cell for TRACE_PAR_COL
    //aux->color = (k == Dd2+1) && (i == 40) ? H-1-j : 0; //column of colored cells for TRACE_PAR_COL
  }
#endif

#ifdef USE_VEGETATION
  /*** Vegetated cells ***/
  //if ((i>=150) && (j==H0)) aux->celltype = VEG;
  //if (/*(i>=150) &&*/ (j>H0)) aux->celltype = VEG;
#endif
#endif // MODEL_DUN
}

///////////////////////////////////////
/// State of a cell 
///////////////////////////////////////
void genesis_cell(Cell *aux, int i, int j, int k)
{
  int id_temp;

  /// default state
  init_cell_default(aux);

  /// templates are applied in reverse order, according to the priority order
  for (id_temp=nb_templates-1; id_temp>=0; id_temp--){
    if (csp_templates[id_temp].type != CSP_CUSTOM)
      init_cell_from_template(aux, i, j, k, id_temp);
    else
      genesis_cell_custom(aux, i, j, k);
  }

  /// final rules
  init_cell_final(aux, i, j, k);
  genesis_cell_final(aux, i, j, k);
}

void genesis()
{
  int i, j, k;
  Cell *aux;

  if ((long)H*L*D > CSP_MAX_CELLS){
    ErrPrintf("ERROR: Too many cells in cellular space (LxHxD=%ld, max is %d). Please use a smaller size.\n",(long)L*H*D, CSP_MAX_CELLS);
    exit(-2);
  }

  HL = H*L;
  HLD = HL*D;

  // allocation of the cellular space
  AllocMemoryPrint("TE",TE, Cell, HLD);
  ResetMemory(TE, Cell, HLD);

  aux = TE;

  // initial setup of the cellular space
  for(k=0; k < D; k++){ // profondeur
    for(j=0; j < H; j++){ // hauteur
      for(i=0; i < L; i++, aux++){ //largeur
        genesis_cell(aux, i, j, k);
      }
    }
  }
}


void dump_terre(int parx, int pary)
{
  FILE *fp;
  char str[256], *filename;
  Cell *pt;
  int i, j;
  char *buf, *aux;
  int dump_type;


  if ((parx == 1) && (pary == 1)){
    dump_type = opt_bin ? DUMP_BIN : DUMP_CSP;
    if (opt_bin)
      sprintf(str,"%s_%d_%d_%d.bin",MOD_NAME,H,L,D);
    else
      sprintf(str,"GENESIS.csp");
    filename = str;
    LogPrintf("format %s\n", opt_bin ? "bin" : "csp");
    csp_set_bounds(0,0,0);
    write_csp(dump_type, filename);
  }
  else{ //mode parallele (raw binary format) //TODO: csp format
    int LX = L / parx;
    int LY = D / pary;
    int tx, ty, tn, k;
    LogPrintf("LX = %d\nLY = %d\n", LX, LY);
    buf = (char*)malloc(sizeof(char)*LX*H);
    for(tn=0, ty=0; ty<pary; ty++){
      for(tx=0; tx<parx; tx++, tn++){
        sprintf(str, "%s_%d_%d_%d_%d.bin", MOD_NAME, H, LX, LY, tn);
        filename = str;
        fp = fopen(filename,"w");
        if (!fp){
          ErrPrintf("erreur creation fichier %s\n", filename);
          exit(-4);
        }
        LogPrintf("creation fichier %s\n", filename);

        for(k=0; k<LY; k++){
          aux = buf;
          pt = TE + (ty*LY+k)*HL + tx*LX;
          for(j=0; j<H; j++, pt+=L){
            for(i=0; i<LX; i++){
              *aux++ = (pt+i)->celltype;
            }
          }
          fwrite(buf, sizeof(char), LX*H, fp);
        }
        fclose(fp);
        //LogPrintf("Data size: %d\n", H*LX*LY);
      }
    }
    free(buf);
  }
}

void dump_par_conf(int parx, int pary)
{
  FILE *fp;
  int i, j, pid;
  char nom[64];

  pid = 0;

  for (j=0; j<pary; j++)
    for (i=0; i<parx; i++, pid++){
      sprintf(nom, "parallel%d.cfg", pid);
      fp = fopen(nom, "w");
      if ( ! fp ){
        ErrPrintf("erreur ouverture fichier parallel.cfg\n");
        exit(-4);
      }
      LogPrintf("creation fichier de configuration %s\n", nom);
      fprintf(fp, "PID %d\n", pid);
      if (!pid) fprintf(fp, "NN %d\n", parx*pary);
      //if (parx > 1){
        if ((i<parx-1) || (boundary == BC_PERIODIC)) fprintf(fp, "EST %d\n", (i<parx-1)? pid+1 : pid+1-parx);
        if ((i>0) || (boundary == BC_PERIODIC)) fprintf(fp, "OUEST %d\n", (i>0)? pid-1 : pid-1+parx);
      //}
      //if (pary > 1){
        if ((j<pary-1) || (boundary == BC_PERIODIC)) fprintf(fp, "SUD %d\n", (j<pary-1)? pid+parx : i);
        if ((j>0) || (boundary == BC_PERIODIC)) fprintf(fp, "NORD %d\n", (j>0)? pid-parx : i+parx*(pary-1));
      //}
      fclose(fp);
    }

}


void params_genesis()
{
  param_family("GENERAL", "General parameters");

  parameter("Use_genesis", "enable genesis (YES|NO)", &use_genesis, PARAM_BOOLEAN, "GENERAL");
  parameter("Model", "name of model", &model_name, PARAM_STRING, "GENERAL");
  parameter("Seed", "random seed - optional", &graine, PARAM_INT, "GENERAL");

  param_family("SPACE", "Space parameters");

  /// dimensions of cellulat space
  parameter("H", "height", &H, PARAM_INT, "SPACE");
  parameter("L", "length", &L, PARAM_INT, "SPACE");
  parameter("D", "depth", &D, PARAM_INT, "SPACE");

  //param_family("GENESIS", "Genesis parameters");
  //parameter("Csp_template", "CSP template for the generation of the initial cellular space", &csp_template_str, PARAM_STRING,"GENESIS");
#ifdef CYCLAGE_HOR
  parameter("Boundary", "boundary conditions (PERIODIC|OPEN|CLOSED|REINJECTION), PERIODIC by default", &boundary_str, PARAM_STRING,"SPACE");
#else
  parameter("Boundary", "boundary conditions (OPEN|CLOSED|REINJECTION), REINJECTION by default", &boundary_str, PARAM_STRING,"SPACE");
#endif
}

void usage()
{
  printf("\nusage: \n genesis L D H [OPTIONS]\n");
  printf(" genesis -f PARAMETERS_FILE [OPTIONS]\n\n");

  param_usage();

  printf("\nCSP TEMPLATES (%d)\n", nb_template_types);
  for(int i=0; i<nb_template_types; i++) printf("  %s\n", t_templates[i].desc);

  printf("\nOPTIONS\n");
  printf("  -bin   \traw binary output\n");
  printf("  -s <n> \trandom seed\n");
#ifdef PARALLEL
  printf("  -par <n>x<m> \tparallel mode, grid <n>x<m>\n");
//  printf("  -node <n> \tnode number <n> in parallel mode\n");
#endif

  exit(0);
}



int main(int argc, char **argv)
{
  int i, n=1, parx = 1, pary = 1;
  unsigned char opt_par = 0;

#ifdef LOG_FILE
  log_file = fopen("GENESIS.log","w");
#endif

  //LogPrintf("genesis\n");
  LogPrintf("genesis %s (%s)\n",VER_NUM,VER_DAT);
  LogPrintf("compiled: %s\n", comp_date);
  LogPrintf("%s model\n",MOD_NAME);

  init_list_params();
  init_template_types();
  params_genesis();
  params_templates();

  if (argc < 3) usage();

  //log the command line options
  LogPrintf("args: ");
  for(i=1; i<argc; i++) LogPrintf("%s ", argv[i]);
  LogPrintf("\n");

  if (!strcmp(argv[n],"-f")){
    char *filename = argv[++n];
    // read parameters
    read_parameters(filename);
    if (!use_genesis){
      LogPrintf("genesis disabled in %s\nquit\n", filename);
      exit(0);
    }
    genesis_parse();
    parse_template_array();
    init_template_array();
  }
  else{
    if (argc < 4) usage();
    L = atoi(argv[n]);
    D = atoi(argv[++n]);
    H = atoi(argv[++n]);
  }

  for(n++; n<argc; n++){
    if (!strcmp(argv[n],"-s") || !strcmp(argv[n],"-g")){
      graine = atoi(argv[++n]);
    }
    else if (!strcmp(argv[n],"-bin")){
      opt_bin = 1;
    }
#ifdef PARALLEL
    else if (!strcmp(argv[n],"-par")){
      sscanf(argv[++n], "%dx%d", &parx, &pary);
      LogPrintf("geom = %s parx = %d \t pary = %d\n", argv[n], parx, pary);
      opt_par = 1;
      L *= parx;
      D *= pary;
    }
/*
    else if (!strcmp(argv[n],"-node")){
      node = atoi(argv[++n]);
    }
*/
#endif
    else{
      ErrPrintf("ERROR: unknown option %s\n", argv[n]);
    }
  }

  LogPrintf("random seed: %d\n", graine);
  srand48(graine);

  if (opt_par) dump_par_conf(parx, pary);

  genesis();

  dump_terre(parx, pary);

  LogPrintf("genesis : done\n");

  return 1;
}


